@@ -37,11 +37,5 @@ extern int petreu_cmdline_options_init(void);
 extern void petreu_init(void);
 extern void petreu_reset(void);
 extern void petreu_shutdown(void);
-extern BYTE read_petreu_reg(WORD addr);
-extern BYTE read_petreu2_reg(WORD addr);
-extern BYTE read_petreu_ram(WORD addr);
-extern void store_petreu_reg(WORD addr, BYTE byte);
-extern void store_petreu2_reg(WORD addr, BYTE byte);
-extern void store_petreu_ram(WORD addr, BYTE byte);
 
 #endif
