@@ -3,6 +3,7 @@
  *
  * Written by
  *  Andreas Boose <viceteam@t-online.de>
+ *  Marco van den Heuvel <blackystardust68@yahoo.com>
  *
  * This file is part of VICE, the Versatile Commodore Emulator.
  * See README for copyright notice.
@@ -31,22 +32,30 @@
 
 #include "attach.h"
 #include "autostart.h"
+#include "bbrtc.h"
+#include "cardkey.h"
+#include "cartio.h"
 #include "cartridge.h"
 #include "clkguard.h"
+#include "coplin_keypad.h"
+#include "cx21.h"
+#include "cx85.h"
 #include "datasette.h"
 #include "debug.h"
+#include "debugcart.h"
 #include "digiblaster.h"
 #include "diskimage.h"
 #include "drive-cmdline-options.h"
 #include "drive-resources.h"
 #include "drive-sound.h"
 #include "drive.h"
-#include "drivecpu.h"
 #include "fliplist.h"
 #include "fsdevice.h"
 #include "gfxoutput.h"
 #include "imagecontents.h"
 #include "init.h"
+#include "joyport.h"
+#include "joystick.h"
 #include "kbdbuf.h"
 #include "keyboard.h"
 #include "log.h"
@@ -58,6 +67,7 @@
 #include "maincpu.h"
 #include "monitor.h"
 #include "network.h"
+#include "paperclip64.h"
 #include "plus4-cmdline-options.h"
 #include "plus4-resources.h"
 #include "plus4-snapshot.h"
@@ -73,6 +83,10 @@
 #include "plus4ui.h"
 #include "printer.h"
 #include "rs232drv.h"
+#include "rushware_keypad.h"
+#include "sampler.h"
+#include "sampler2bit.h"
+#include "sampler4bit.h"
 #include "screenshot.h"
 #include "serial.h"
 #include "sid.h"
@@ -81,42 +95,66 @@
 #include "sid-resources.h"
 #include "sound.h"
 #include "tape.h"
+#include "tapeport.h"
 #include "ted-cmdline-options.h"
 #include "ted-resources.h"
 #include "ted-sound.h"
 #include "ted.h"
 #include "traps.h"
 #include "types.h"
+#include "userport.h"
+#include "userport_dac.h"
+#include "userport_joystick.h"
 #include "vice-event.h"
 #include "video.h"
 #include "video-sound.h"
 #include "vsync.h"
 
+#ifdef HAVE_MOUSE
+#include "mouse.h"
+#endif
+
+
+/** \brief  Delay in seconds before pasting -keybuf argument into the buffer
+ */
+#define KBDBUF_ALARM_DELAY   1
+
+
 machine_context_t machine_context;
 
-#define NUM_KEYBOARD_MAPPINGS 2
-
-static char *machine_keymap_res_name_list[NUM_KEYBOARD_MAPPINGS] = {
-    "KeymapSymFile", "KeymapPosFile"
-};
-
-char *machine_keymap_file_list[NUM_KEYBOARD_MAPPINGS] = {
-    NULL, NULL
-};
-
-char *machine_get_keymap_res_name(int val)
-{
-    if (val < 0 || val > NUM_KEYBOARD_MAPPINGS) {
-        return NULL;
-    }
-    return machine_keymap_res_name_list[val];
-}
-
 const char machine_name[] = "PLUS4";
 int machine_class = VICE_MACHINE_PLUS4;
 
 static void machine_vsync_hook(void);
 
+int machine_get_keyboard_type(void)
+{
+#if 0
+    int type;
+    if (resources_get_int("KeyboardType", &type) < 0) {
+        return 0;
+    }
+    return type;
+#endif
+    return 0;
+}
+
+char *machine_get_keyboard_type_name(int type)
+{
+    return NULL; /* return 0 if no different types exist */
+}
+
+/* return number of available keyboard types for this machine */
+int machine_get_num_keyboard_types(void)
+{
+    return 1;
+}
+
+kbdtype_info_t *machine_get_keyboard_info_list(void)
+{
+    return NULL; /* return 0 if no different types exist */
+}
+
 /* ------------------------------------------------------------------------- */
 
 static const trap_t plus4_serial_traps[] = {
@@ -255,6 +293,63 @@ static long cycles_per_rfsh = PLUS4_PAL_CYCLES_PER_RFSH;
 static double rfsh_per_sec = PLUS4_PAL_RFSH_PER_SEC;
 */
 
+static joyport_port_props_t control_port_1 =
+{
+    "Control port 1",
+    0,                  /* has NO potentiometer connected to this port */
+    0,                  /* has NO lightpen support on this port */
+    1                   /* port is always active */
+};
+
+static joyport_port_props_t control_port_2 =
+{
+    "Control port 2",
+    0,                  /* has NO potentiometer connected to this port */
+    0,                  /* has NO lightpen support on this port */
+    1                   /* port is always active */
+};
+
+static joyport_port_props_t userport_joy_control_port_1 =
+{
+    "Userport joystick adapter port 1",
+    0,                  /* has NO potentiometer connected to this port */
+    0,                  /* has NO lightpen support on this port */
+    0                   /* port can be switched on/off */
+};
+
+static joyport_port_props_t userport_joy_control_port_2 =
+{
+    "Userport joystick adapter port 2",
+    0,                  /* has NO potentiometer connected to this port */
+    0,                  /* has NO lightpen support on this port */
+    0                   /* port can be switched on/off */
+};
+
+static joyport_port_props_t sidcard_port =
+{
+    "SIDCard control port",
+    1,                  /* has a potentiometer connected to this port */
+    0,                  /* has NO lightpen support on this port */
+    0                   /* port can be switched on/off */
+};
+
+static int init_joyport_ports(void)
+{
+    if (joyport_port_register(JOYPORT_1, &control_port_1) < 0) {
+        return -1;
+    }
+    if (joyport_port_register(JOYPORT_3, &userport_joy_control_port_1) < 0) {
+        return -1;
+    }
+    if (joyport_port_register(JOYPORT_4, &userport_joy_control_port_2) < 0) {
+        return -1;
+    }
+    if (joyport_port_register(JOYPORT_5, &sidcard_port) < 0) {
+        return -1;
+    }
+    return joyport_port_register(JOYPORT_2, &control_port_2);
+}
+
 /* Plus4-specific resource initialization.  This is called before initializing
    the machine itself with `machine_init()'.  */
 int machine_resources_init(void)
@@ -271,6 +366,10 @@ int machine_resources_init(void)
         init_resource_fail("ted");
         return -1;
     }
+    if (cartio_resources_init() < 0) {
+        init_resource_fail("cartio");
+        return -1;
+    }
     if (cartridge_resources_init() < 0) {
         init_resource_fail("cartridge");
         return -1;
@@ -310,14 +409,74 @@ int machine_resources_init(void)
         return -1;
     }
 #endif
+    if (init_joyport_ports() < 0) {
+        init_resource_fail("joyport ports");
+        return -1;
+    }
+    if (joyport_resources_init() < 0) {
+        init_resource_fail("joyport devices");
+        return -1;
+    }
+    if (joyport_sampler2bit_resources_init() < 0) {
+        init_resource_fail("joyport 2bit sampler");
+        return -1;
+    }
+    if (joyport_sampler4bit_resources_init() < 0) {
+        init_resource_fail("joyport 4bit sampler");
+        return -1;
+    }
+    if (joyport_bbrtc_resources_init() < 0) {
+        init_resource_fail("joyport bbrtc");
+        return -1;
+    }
+    if (joyport_paperclip64_resources_init() < 0) {
+        init_resource_fail("joyport paperclip64 dongle");
+        return -1;
+    }
+    if (joyport_coplin_keypad_resources_init() < 0) {
+        init_resource_fail("joyport coplin keypad");
+        return -1;
+    }
+    if (joyport_cx21_resources_init() < 0) {
+        init_resource_fail("joyport cx21 keypad");
+        return -1;
+    }
+    if (joyport_cx85_resources_init() < 0) {
+        init_resource_fail("joyport cx85 keypad");
+        return -1;
+    }
+    if (joyport_rushware_keypad_resources_init() < 0) {
+        init_resource_fail("joyport rushware keypad");
+        return -1;
+    }
+    if (joyport_cardkey_resources_init() < 0) {
+        init_resource_fail("joyport cardkey keypad");
+        return -1;
+    }
     if (joystick_resources_init() < 0) {
         init_resource_fail("joystick");
         return -1;
     }
+    if (userport_resources_init() < 0) {
+        init_resource_fail("userport devices");
+        return -1;
+    }
+    if (userport_joystick_resources_init() < 0) {
+        init_resource_fail("userport joystick");
+        return -1;
+    }
+    if (userport_dac_resources_init() < 0) {
+        init_resource_fail("userport dac");
+        return -1;
+    }
     if (gfxoutput_resources_init() < 0) {
         init_resource_fail("gfxoutput");
         return -1;
     }
+    if (sampler_resources_init() < 0) {
+        init_resource_fail("samplerdrv");
+        return -1;
+    }
     if (fliplist_resources_init() < 0) {
         init_resource_fail("flip list");
         return -1;
@@ -359,9 +518,9 @@ int machine_resources_init(void)
         return -1;
     }
 #endif
-#ifndef COMMON_KBD
-    if (kbd_resources_init() < 0) {
-        init_resource_fail("kbd");
+#ifdef HAVE_MOUSE
+    if (mouse_resources_init() < 0) {
+        init_resource_fail("mouse");
         return -1;
     }
 #endif
@@ -369,10 +528,23 @@ int machine_resources_init(void)
         init_resource_fail("drive");
         return -1;
     }
+    /*
+     * This needs to be called before tapeport_resources_init(), otherwise
+     * the tapecart will fail to initialize due to the Datasette resource
+     * appearing after the Tapecart resources
+     */
     if (datasette_resources_init() < 0) {
         init_resource_fail("datasette");
         return -1;
     }
+    if (tapeport_resources_init() < 0) {
+        init_resource_fail("tapeport");
+        return -1;
+    }
+    if (debugcart_resources_init() < 0) {
+        init_resource_fail("debug cart");
+        return -1;
+    }
     return 0;
 }
 
@@ -387,6 +559,12 @@ void machine_resources_shutdown(void)
     drive_resources_shutdown();
     fsdevice_resources_shutdown();
     disk_image_resources_shutdown();
+    sampler_resources_shutdown();
+    cartio_shutdown();
+    userport_resources_shutdown();
+    joyport_bbrtc_resources_shutdown();
+    tapeport_resources_shutdown();
+    debugcart_resources_shutdown();
 }
 
 /* Plus4-specific command-line option initialization.  */
@@ -404,6 +582,10 @@ int machine_cmdline_options_init(void)
         init_cmdline_options_fail("ted");
         return -1;
     }
+    if (cartio_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("cartio");
+        return -1;
+    }
     if (cartridge_cmdline_options_init() < 0) {
         init_cmdline_options_fail("cartridge");
         return -1;
@@ -443,14 +625,38 @@ int machine_cmdline_options_init(void)
         return -1;
     }
 #endif
+    if (joyport_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("joyport");
+        return -1;
+    }
+    if (joyport_bbrtc_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("bbrtc");
+        return -1;
+    }
     if (joystick_cmdline_options_init() < 0) {
         init_cmdline_options_fail("joystick");
         return -1;
     }
+    if (userport_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("userport");
+        return -1;
+    }
+    if (userport_joystick_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("userport joystick");
+        return -1;
+    }
+    if (userport_dac_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("userport dac");
+        return -1;
+    }
     if (gfxoutput_cmdline_options_init() < 0) {
         init_cmdline_options_fail("gfxoutput");
         return -1;
     }
+    if (sampler_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("samplerdrv");
+        return -1;
+    }
     if (fliplist_cmdline_options_init() < 0) {
         init_cmdline_options_fail("flip list");
         return -1;
@@ -479,15 +685,21 @@ int machine_cmdline_options_init(void)
         init_cmdline_options_fail("autostart");
         return -1;
     }
+#ifdef HAVE_NETWORK
+    if (network_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("network");
+        return -1;
+    }
+#endif
 #ifdef DEBUG
     if (debug_cmdline_options_init() < 0) {
         init_cmdline_options_fail("debug");
         return -1;
     }
 #endif
-#ifndef COMMON_KBD
-    if (kbd_cmdline_options_init() < 0) {
-        init_cmdline_options_fail("kbd");
+#ifdef HAVE_MOUSE
+    if (mouse_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("mouse");
         return -1;
     }
 #endif
@@ -495,10 +707,18 @@ int machine_cmdline_options_init(void)
         init_cmdline_options_fail("drive");
         return -1;
     }
+    if (tapeport_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("tapeport");
+        return -1;
+    }
     if (datasette_cmdline_options_init() < 0) {
         init_cmdline_options_fail("datasette");
         return -1;
     }
+    if (debugcart_cmdline_options_init() < 0) {
+        init_cmdline_options_fail("debug cart");
+        return -1;
+    }
     return 0;
 }
 
@@ -517,7 +737,7 @@ static void plus4_monitor_init(void)
     asmR65C02_init(&asmR65C02);
 
     for (dnr = 0; dnr < DRIVE_NUM; dnr++) {
-        drive_interface_init[dnr] = drivecpu_monitor_interface_get(dnr);
+        drive_interface_init[dnr] = drive_cpu_monitor_interface_get(dnr);
     }
 
     /* Initialize the monitor.  */
@@ -547,12 +767,13 @@ int machine_specific_init(void)
     /* Setup trap handling.  */
     traps_init();
 
-    if (!video_disabled_mode) {
-        joystick_init();
-    }
-
     gfxoutput_init();
 
+#ifdef HAVE_MOUSE
+    /* Initialize mouse support (if present).  */
+    mouse_init();
+#endif
+
     /* Initialize serial traps.  */
     if (serial_init(plus4_serial_traps) < 0) {
         return -1;
@@ -582,7 +803,7 @@ int machine_specific_init(void)
     if (delay == 0) {
         delay = 2; /* default */
     }
-    autostart_init((CLOCK)(delay * PLUS4_PAL_RFSH_PER_SEC * PLUS4_PAL_CYCLES_PER_RFSH), 0, 0, 0xc8, 0xca, -40);
+    autostart_init((CLOCK)(delay * PLUS4_PAL_RFSH_PER_SEC * PLUS4_PAL_CYCLES_PER_RFSH), 1, 0, 0xc8, 0xca, -40);
 
     /* Initialize the sidcart first */
     sidcart_sound_chip_init();
@@ -594,6 +815,9 @@ int machine_specific_init(void)
     digiblaster_sound_chip_init();
     speech_sound_chip_init();
 
+    /* Initialize userport based sound chips */
+    userport_dac_sound_chip_init();
+
     drive_sound_init();
     video_sound_init();
 
@@ -601,18 +825,18 @@ int machine_specific_init(void)
        device yet.  */
     sound_init(machine_timing.cycles_per_sec, machine_timing.cycles_per_rfsh);
 
+    /* Pre-init Plus4-specific parts of the menus before ted_init()
+       creates a canvas window with a menubar at the top. */
+    if (!console_mode) {
+        plus4ui_init_early();
+    }
+
     if (ted_init() == NULL) {
         return -1;
     }
 
     acia_init();
 
-#ifndef COMMON_KBD
-    if (plus4_kbd_init() < 0) {
-        return -1;
-    }
-#endif
-
     plus4_monitor_init();
 
     /* Initialize vsync and register our hook function.  */
@@ -621,9 +845,16 @@ int machine_specific_init(void)
                                 machine_timing.cycles_per_sec);
 
     /* Initialize keyboard buffer.  */
-    kbdbuf_init(1319, 239, 8, (CLOCK)(machine_timing.rfsh_per_sec * machine_timing.cycles_per_rfsh));
+    kbdbuf_init(1319, 239, 8, (CLOCK)(machine_timing.rfsh_per_sec *
+                machine_timing.cycles_per_rfsh * KBDBUF_ALARM_DELAY));
 
-    plus4ui_init();
+    if (!console_mode) {
+        plus4ui_init();
+    }
+
+    if (!video_disabled_mode) {
+        joystick_init();
+    }
 
     cs256k_init();
 
@@ -633,17 +864,9 @@ int machine_specific_init(void)
 
     machine_drive_stub();
 
-#if defined (USE_XF86_EXTENSIONS) && \
-    (defined(USE_XF86_VIDMODE_EXT) || defined (HAVE_XRANDR))
-    {
-        /* set fullscreen if user used `-fullscreen' on cmdline */
-        int fs;
-        resources_get_int("UseFullscreen", &fs);
-        if (fs) {
-            resources_set_int("TEDFullscreen", 1);
-        }
-    }
-#endif
+    /* Initialize the machine specific I/O */
+    plus4io_init();
+
     return 0;
 }
 
@@ -669,6 +892,8 @@ void machine_specific_reset(void)
 
     drive_reset();
     datasette_reset();
+
+    sampler_reset();
 }
 
 void machine_specific_powerup(void)
@@ -686,7 +911,13 @@ void machine_specific_shutdown(void)
     cs256k_shutdown();
     h256k_shutdown();
 
-    plus4ui_shutdown();
+#ifdef HAVE_MOUSE
+    mouse_shutdown();
+#endif
+
+    if (!console_mode) {
+        plus4ui_shutdown();
+    }
 }
 
 void machine_handle_pending_alarms(int num_write_cycles)
@@ -711,7 +942,7 @@ static void machine_vsync_hook(void)
 
     /* The drive has to deal both with our overflowing and its own one, so
        it is called even when there is no overflowing in the main CPU.  */
-    drivecpu_prevent_clk_overflow_all(sub);
+    drive_cpu_prevent_clk_overflow_all(sub);
 }
 
 void machine_set_restore_key(int v)
@@ -744,34 +975,8 @@ void machine_get_line_cycle(unsigned int *line, unsigned int *cycle, int *half_c
     *half_cycle = (int)-1;
 }
 
-void machine_change_timing(int timeval)
+void machine_change_timing(int timeval, int border_mode)
 {
-    int border_mode;
-
-    switch (timeval) {
-        default:
-        case MACHINE_SYNC_PAL ^ TED_BORDER_MODE(TED_NORMAL_BORDERS):
-        case MACHINE_SYNC_NTSC ^ TED_BORDER_MODE(TED_NORMAL_BORDERS):
-            timeval ^= TED_BORDER_MODE(TED_NORMAL_BORDERS);
-            border_mode = TED_NORMAL_BORDERS;
-            break;
-        case MACHINE_SYNC_PAL ^ TED_BORDER_MODE(TED_FULL_BORDERS):
-        case MACHINE_SYNC_NTSC ^ TED_BORDER_MODE(TED_FULL_BORDERS):
-            timeval ^= TED_BORDER_MODE(TED_FULL_BORDERS);
-            border_mode = TED_FULL_BORDERS;
-            break;
-        case MACHINE_SYNC_PAL ^ TED_BORDER_MODE(TED_DEBUG_BORDERS):
-        case MACHINE_SYNC_NTSC ^ TED_BORDER_MODE(TED_DEBUG_BORDERS):
-            timeval ^= TED_BORDER_MODE(TED_DEBUG_BORDERS);
-            border_mode = TED_DEBUG_BORDERS;
-            break;
-        case MACHINE_SYNC_PAL ^ TED_BORDER_MODE(TED_NO_BORDERS):
-        case MACHINE_SYNC_NTSC ^ TED_BORDER_MODE(TED_NO_BORDERS):
-            timeval ^= TED_BORDER_MODE(TED_NO_BORDERS);
-            border_mode = TED_NO_BORDERS;
-            break;
-    }
-
     switch (timeval) {
         case MACHINE_SYNC_PAL:
             machine_timing.cycles_per_sec = PLUS4_PAL_CYCLES_PER_SEC;
@@ -801,6 +1006,9 @@ void machine_change_timing(int timeval)
                                 machine_timing.screen_lines);
     drive_set_machine_parameter(machine_timing.cycles_per_sec);
     serial_iec_device_set_machine_parameter(machine_timing.cycles_per_sec);
+#ifdef HAVE_MOUSE
+    neos_mouse_set_machine_parameter(machine_timing.cycles_per_sec);
+#endif
     clk_guard_set_clk_base(maincpu_clk_guard, machine_timing.cycles_per_rfsh);
 
     ted_change_timing(&machine_timing, border_mode);
@@ -851,28 +1059,49 @@ int machine_canvas_async_refresh(struct canvas_refresh_s *refresh,
     return 0;
 }
 
-int machine_num_keyboard_mappings(void)
-{
-    return NUM_KEYBOARD_MAPPINGS;
-}
-
 struct image_contents_s *machine_diskcontents_bus_read(unsigned int unit)
 {
     return diskcontents_iec_read(unit);
 }
 
-BYTE machine_tape_type_default(void)
+uint8_t machine_tape_type_default(void)
 {
-    return TAPE_CAS_TYPE_BAS;
+    return TAPE_CAS_TYPE_PRG;
+}
+
+uint8_t machine_tape_behaviour(void)
+{
+    return TAPE_BEHAVIOUR_C16;
 }
 
 int machine_addr_in_ram(unsigned int addr)
 {
     /* FIXME are these correct? */
-    return (addr < 0xe000 && !(addr >= 0xa000 && addr < 0xc000)) ? 1 : 0;
+    /* ROM is 8000-bfff
+     *        d000-fcff
+     *        fd00-     I/O
+     */
+    return ((addr < 0x8000) || (addr >= 0xc000 && addr <= 0xcfff)) ? 1 : 0;
 }
 
 const char *machine_get_name(void)
 {
     return machine_name;
 }
+
+/* ------------------------------------------------------------------------- */
+
+static userport_port_props_t userport_props = {
+    0, /* NO pa2 pin */
+    0, /* NO pa3 pin */
+    NULL, /* NO flag pin */
+    0, /* NO pc pin */
+    0  /* NO cnt1, cnt2 or sp pins */
+};
+
+int machine_register_userport(void)
+{
+    userport_port_register(&userport_props);
+
+    return 0;
+}
