@@ -30,22 +30,25 @@
 #include <string.h>
 
 #include "cmdline.h"
+#include "lib.h"
+#include "machine.h"
 #include "ram.h"
 #include "resources.h"
-#include "translate.h"
 #include "types.h"
 
-static int start_value;
-static int value_invert;
-static int pattern_invert;
+static int start_value = 0;
+static int value_invert = 64;
+static int pattern_invert = 0;
 
 static int set_start_value(int val, void *param)
 {
     start_value = val;
-    if (start_value < 0)
+    if (start_value < 0) {
         start_value = 0;
-    if (start_value > 0xff)
+    }
+    if (start_value > 0xff) {
         start_value = 0xff;
+    }
     return 0;
 }
 
@@ -70,44 +73,44 @@ static const resource_int_t resources_int[] = {
       &value_invert, set_value_invert, NULL },
     { "RAMInitPatternInvert", 0, RES_EVENT_SAME, NULL,
       &pattern_invert, set_pattern_invert, NULL },
-    { NULL }
+    RESOURCE_INT_LIST_END
 };
 
 int ram_resources_init(void)
 {
-    return resources_register_int(resources_int);
+    if (machine_class != VICE_MACHINE_VSID) {
+        return resources_register_int(resources_int);
+    }
+    return 0;
 }
 
-static const cmdline_option_t cmdline_options[] = {
-    { "-raminitstartvalue", SET_RESOURCE, 1,
+static const cmdline_option_t cmdline_options[] =
+{
+    { "-raminitstartvalue", SET_RESOURCE, CMDLINE_ATTRIB_NEED_ARGS,
       NULL, NULL, "RAMInitStartValue", NULL,
-      USE_PARAM_ID, USE_DESCRIPTION_ID,
-      IDCLS_P_VALUE, IDCLS_SET_FIRST_RAM_ADDRESS_VALUE,
-      NULL, NULL },
-    { "-raminitvalueinvert" , SET_RESOURCE, 1,
+      "<value>", "Set the value for the very first RAM address after powerup" },
+    { "-raminitvalueinvert", SET_RESOURCE, CMDLINE_ATTRIB_NEED_ARGS,
       NULL, NULL, "RAMInitValueInvert", NULL,
-      USE_PARAM_ID, USE_DESCRIPTION_ID,
-      IDCLS_P_NUM_OF_BYTES, IDCLS_LENGTH_BLOCK_SAME_VALUE,
-      NULL, NULL },
-    { "-raminitpatterninvert", SET_RESOURCE, 1,
+      "<num of bytes>", "Length of memory block initialized with the same value" },
+    { "-raminitpatterninvert", SET_RESOURCE, CMDLINE_ATTRIB_NEED_ARGS,
       NULL, NULL, "RAMInitPatternInvert", NULL,
-      USE_PARAM_ID, USE_DESCRIPTION_ID,
-      IDCLS_P_NUM_OF_BYTES, IDCLS_LENGTH_BLOCK_SAME_PATTERN,
-      NULL, NULL },
-    { NULL }
+      "<num of bytes>", "Length of memory block initialized with the same pattern" },
+    CMDLINE_LIST_END
 };
 
 int ram_cmdline_options_init(void)
 {
-    return cmdline_register_options(cmdline_options);
+    if (machine_class != VICE_MACHINE_VSID) {
+        return cmdline_register_options(cmdline_options);
+    }
+    return 0;
 }
 
 
-void ram_init(BYTE *memram, unsigned int ramsize)
+void ram_init(uint8_t *memram, unsigned int ramsize)
 {
-
     unsigned int i, j, k, l;
-    BYTE v = start_value;
+    uint8_t v = start_value;
 
     j = value_invert - 1;
     k = pattern_invert - 1;
@@ -132,60 +135,45 @@ void ram_init(BYTE *memram, unsigned int ramsize)
     }
 }
 
-
-const char *ram_init_print_pattern(void)
+/* create a preview of the RAM init pattern - this should be as fast as
+   possible since it is used in the GUI */
+void ram_init_print_pattern(char *s, int len, char *eol)
 {
-    static char s[512], s_tmp[16], pattern_line[64];
-    BYTE v = start_value;
-    int i;
-    int linenum = 0;
-    int last_line_drawn = 0;
+    unsigned char *mem;
+    char *p = s, *pp;
+    int i, a;
+    const char hextab[16] = "0123456789abcdef";
 
-    s[0] = 0;
+    mem = lib_malloc(len);
+    ram_init(mem, len);
 
-    do {
-        pattern_line[0] = 0;
-
-        for (i = 0; i < 8; i++) {
-            sprintf(s_tmp," %02x", v);
-            
-            strcat(pattern_line, s_tmp);
-
-            if (value_invert > 0
-                && (linenum * 8 + i + 1) % value_invert == 0)
-            {
-                v ^= 0xff;
-            }
-
-            if (pattern_invert > 0
-                && (linenum * 8 + i + 1) % pattern_invert == 0)
-            {
-                v ^= 0xff;
+    for (a = 0; a < len;) {
+        /* add address at start if line */
+        *p++ = hextab[(a >> 12) & 0x0f];
+        *p++ = hextab[(a >> 8) & 0x0f];
+        *p++ = hextab[(a >> 4) & 0x0f];
+        *p++ = hextab[a & 0x0f];
+        *p++ = ':';
+        *p++ = ' ';
+        /* add 16 bytes hex dump */
+        for (i = 0; i < 16; i++, a++) {
+            *p++ = hextab[(mem[a] >> 4) & 0x0f];
+            *p++ = hextab[(mem[a]) & 0x0f];
+            *p++ = ' ';
+        }
+        /* add end of line */
+        pp = eol;
+        while (*pp) {
+            *p++ = *pp++;
+        }
+        /* after each full page add another end of line */
+        if ((a & 0xff) == 0) {
+            pp = eol;
+            while (*pp) {
+                *p++ = *pp++;
             }
         }
-
-        if (linenum * 8 == 0
-            || linenum * 8 == value_invert
-            || linenum * 8 == pattern_invert
-            || linenum * 8 == pattern_invert + value_invert)
-        {
-            sprintf(s_tmp, "%04x ", linenum * 8);
-            strcat(s, s_tmp);
-            strcat(s, pattern_line);
-            strcat(s, "\n");
-            last_line_drawn = 1;
-        } else {
-            if (last_line_drawn == 1)
-                strcat(s, "...\n");
-            last_line_drawn = 0;
-        }
-
-        linenum++;
-
-    } while (linenum * 8 < value_invert * 2 || linenum * 8 < pattern_invert * 2);
-
-    if (last_line_drawn == 1)
-        strcat(s, "...\n");
-
-    return s;
+    }
+    *p = 0;
+    lib_free(mem);
 }
